/* 
   COPYRIGHT 2006 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - win32_fport.c */
/* rdk@krupczak.org */

/* port of fport.c/portuser.c/portuser.cpp originally written by flashkey 
   arkadyf@hotmail.com 

   fixes for leaking handles and memory by rdk 6/14/2005

   code uses undocumented Native API calls; the book 
   "Windows NT/2000 Native API Reference" does document some of the
   APIs 
*/

#ifdef WIN32 

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <Aclapi.h>
#include <winioctl.h>
#include <pthread.h>
#pragma comment ( lib, "ws2_32.lib" )

#include "xmp.h"
#include "plugins.h"
#include "mib2plugin.h"
#include "xmp_vars.h"

#define false 0
#define true 1

/* this code retrieves a mapping of pid to port but not necessarily
   pid to connection or socket; subsequently, we have to be
   careful that we map the proper pid/port combo so that we
   dont mix up endpoints 
  
   this condition can show up if two PIDs each have a socket open
   with same port/proto (e.g. if SO_REUSEADDR is enabled)

*/

/* ************************************************************* */
/* structures defined here rather than included so we dont have to 
   have every sdk/ddk installed */

typedef struct _HandleInfo
{
  USHORT dwPid;
  USHORT CreatorBackTraceIndex;
  BYTE   ObjType;
  BYTE   HandleAttributes;
  USHORT HndlOffset;
  DWORD  dwKeObject;
  ULONG  GrantedAccess;
} HANDLEINFO, *PHANDLEINFO;

typedef struct _IO_STATUS_BLOCK {
  DWORD Status;
  ULONG Information;
} IO_STATUS_BLOCK, *PIO_STATUS_BLOCK;

typedef struct _LSA_UNICODE_STRING {
    USHORT Length;
    USHORT MaximumLength;
    PWSTR  Buffer;
} LSA_UNICODE_STRING, *PLSA_UNICODE_STRING;

typedef LSA_UNICODE_STRING UNICODE_STRING, *PUNICODE_STRING;

typedef struct _OBJECT_ATTRIBUTES {
  ULONG Length;
  HANDLE RootDirectory;
  UNICODE_STRING *ObjectName;
  ULONG Attributes;
  PSECURITY_DESCRIPTOR SecurityDescriptor;
  PSECURITY_QUALITY_OF_SERVICE SecurityQualityOfService;
} OBJECT_ATTRIBUTES,*POBJECT_ATTRIBUTES;

typedef struct _handleMap {
  pid_t pid;
  u_int port;
  u_int proto;      /* tcp(6), udp(17) */
} HandleMap_t;

static HandleMap_t *handleCache = NULL;
static int numHandles = 0, maxHandles = 0;

/* ************************************************************* */
/* function prototypes defined and then populated at runtime */

typedef DWORD (CALLBACK* NTQUERYSYSTEMINFORMATION)( DWORD, PDWORD, DWORD, 
PVOID );
static NTQUERYSYSTEMINFORMATION NtQuerySystemInformation = NULL;

typedef VOID (CALLBACK* RTLINITUNICODESTRING)(PUNICODE_STRING,PCWSTR);
static RTLINITUNICODESTRING RtlInitUnicodeString = NULL;

typedef DWORD (CALLBACK* ZWOPENSECTION)(PVOID, DWORD,POBJECT_ATTRIBUTES);
static ZWOPENSECTION ZwOpenSection = NULL;

/* this function and symbol are apparently unused */

typedef VOID(CALLBACK* 
INITIALIZEOBJECTATTRIBUTES)(POBJECT_ATTRIBUTES,PUNICODE_STRING,ULONG,HANDLE,PSECURITY_DESCRIPTOR);
static INITIALIZEOBJECTATTRIBUTES InitializeObjectAttributes = NULL;

typedef DWORD (CALLBACK* 
ZWOPENFILE)(PHANDLE,DWORD,POBJECT_ATTRIBUTES,PIO_STATUS_BLOCK,ULONG,ULONG);
static ZWOPENFILE ZwOpenFile = NULL;

typedef DWORD (CALLBACK* ZWCLOSE)(HANDLE);
static ZWCLOSE ZwClose = NULL;

/* ************************************************************* */

static int enlarge_handle_cache()
{
     /* if need initializing */
     if (handleCache == NULL) {
        handleCache = (HandleMap_t *)malloc(sizeof(HandleMap_t) * 50);
        if (handleCache == NULL) {
           perror("enlarge_handle_cache: initial malloc failed");
           return -1;     
        }
        maxHandles = 50;
      }

     if (maxHandles == numHandles) {
        maxHandles += 25;
        handleCache = (HandleMap_t *)realloc(handleCache,
					     maxHandles*sizeof(HandleMap_t));
        if (handleCache == NULL) {
	   maxHandles = numHandles = 0;
           return -1;
        }
     }

     return 1;
}

/* ************************************************************* */
/* Given a port/proto combination, lookup and return corresponding
   PID; once we find a match, zero out entry so that it is
   not returned again.  Next time the cache is re-generated,
   we start back over; we assume these functions are all
   called from single mib2 worker thread and that locking is
   not necessary because actual XMP queries never invoke these
   functions 
*/
  
pid_t lookupProcessID(u_int port, u_int proto)
{
     int i;

     for (i=0; i<numHandles; i++) {

         if ((port == handleCache[i].port) &&
	     (proto == handleCache[i].proto)) {

	    handleCache[i].port = 0;
            handleCache[i].proto = 0;
            return handleCache[i].pid;
         }
     }
 
     return (pid_t)-1;
}

/* ************************************************************* */

static DWORD getmap(PHANDLEINFO get1,LPVOID addr,HANDLE pm,char * buf)
{
     DWORD readset;
     LPVOID pmaddr1;
     int i;

     readset = (get1->dwKeObject>>0x16);
     readset = *((LPDWORD)((DWORD)addr + 4*readset));

     if ((readset&0x000000ff)<1)
        return 0;
     if((readset&0x000000ff)<0x80) {
        pmaddr1 = MapViewOfFile(pm,4,0,readset&0xfffff000,0x1000);
//		if(!pmaddr1)
//			return 0;
	readset = (get1->dwKeObject>>0x0c) & 0x3ff;
	readset = *((LPDWORD)((DWORD)pmaddr1 + 4*readset));
	UnmapViewOfFile(pmaddr1);
	readset = readset & 0x0FFFFF000;
     }
     else
	readset=(readset&0xfffff000)+(get1->dwKeObject&0x003ff000);

     pmaddr1 = MapViewOfFile(pm,4,0,readset,0x1000);
     if (pmaddr1!=NULL)	{
        readset = get1->dwKeObject&0x00000fff;
	readset = (DWORD)pmaddr1+readset;
	for(i=0;i<0x70;i++)
	   buf[i] = *((char *)(readset + i));
	UnmapViewOfFile(pmaddr1);
     }
     else
        return 0;
    return readset;
}

/* ************************************************************* */

#define NT_HANDLE_LIST 16
#define MAX_HANDLE_LIST_BUF 0x200000

static DWORD dwNumBytes = MAX_HANDLE_LIST_BUF;
static PDWORD pdwHandleList = NULL;
/* static PDWORD pdwHandInfo = NULL; */

int win2kfport( )
{
     DWORD readset1;
     DWORD readset2;
     DWORD readset3;
     OVERLAPPED la;
     static HMODULE hNtdll = NULL;
     DWORD dwNumEntries;
     PHANDLEINFO pHandleInfo;
     HANDLE htcp,hudp;
     HANDLE pmy;
     HANDLE myhand;
     HANDLE h1=NULL;
     DWORD status;
     LPVOID pmaddr;
     TOKEN_PRIVILEGES NewState;
     DWORD dwNumBytesRet;
     HANDLE hToken;
     BOOL isok;
     UNICODE_STRING dn;
     IO_STATUS_BLOCK ch3;
     int port1;
     int port2;
     HANDLE hProc;
     wchar_t * ch1 = L"\\Device\\Tcp";
     wchar_t * ch2 = L"\\Device\\Udp";
     OBJECT_ATTRIBUTES ofs;
     DWORD i;
     DWORD p=0;
     char buf1[0x70];
     char buf2[0x70];
     char buf3[0x70];
     char in[0x18];
     char in1[0x18];
     char out[0x38];
     char out1[0x30];
     PHANDLEINFO tcpdnum;
     PHANDLEINFO udpdnum;
     OBJECT_ATTRIBUTES udm= {
		sizeof(OBJECT_ATTRIBUTES),// Length
		NULL, // RootDirectory
		&dn,  // ObjectName
		0,   // Attributes
		NULL, // SecurityDescriptor
		NULL, // SecurityQualityOfService
     };

     /* first time we're called, load ntdll, get some function
        pointers, etc. */

     if (hNtdll == NULL) {
        hNtdll = LoadLibrary( "ntdll.dll" );
        if ( !hNtdll ) {
           printf( "LoadLibrary( NTDLL.DLL ) Error:%d\n", GetLastError() );
           return -1;
        }

        /* get address of various functions */
        NtQuerySystemInformation = (NTQUERYSYSTEMINFORMATION)
            GetProcAddress( hNtdll, "NtQuerySystemInformation");
        RtlInitUnicodeString = (RTLINITUNICODESTRING)
           GetProcAddress( hNtdll, "RtlInitUnicodeString");
        ZwOpenSection = (ZWOPENSECTION)
           GetProcAddress( hNtdll, "ZwOpenSection");;
        ZwOpenFile = (ZWOPENFILE)
           GetProcAddress( hNtdll, "ZwOpenFile");
        ZwClose = (ZWCLOSE)
           GetProcAddress( hNtdll, "ZwClose");

        InitializeObjectAttributes = (INITIALIZEOBJECTATTRIBUTES)
           GetProcAddress( hNtdll, "InitializeObjectAttributes");

        if ( (NtQuerySystemInformation == NULL) ||
	     (RtlInitUnicodeString == NULL) ||
	     (ZwOpenSection == NULL) ||
	     (ZwOpenFile == NULL) ||
	     (ZwClose == NULL) ) {
	   printf("win2kfport: failed to find all required symbols\n");
           return -1;
        }

        // malloc space once for handles; 
	pdwHandleList = (PDWORD)malloc(dwNumBytes);
	/* pdwHandInfo = (PDWORD)malloc(2048); */

        if (pdwHandleList == NULL) {
	   printf("win2kfport: failed to alloc handle list\n");
           FreeLibrary(hNtdll);
           hNtdll = NULL;
           return -1;
        }
 
     } /* if hNtdll == null */

     RtlInitUnicodeString(&dn,L"\\Device\\PhysicalMemory");
     status = ZwOpenSection(&h1,SECTION_MAP_READ,&udm);

     if (status == 0) {
	pmy = GetCurrentProcess();
	pmaddr = MapViewOfFile(h1,4,0,0x30000,0x1000);
	NewState.PrivilegeCount=1;
	NewState.Privileges[0].Attributes=2;
	NewState.Privileges[0].Luid.HighPart=0;
	NewState.Privileges[0].Luid.LowPart=0;

	isok=LookupPrivilegeValue(0,SE_DEBUG_NAME,
                                    &NewState.Privileges[0].Luid);

	isok=OpenProcessToken(pmy,0x20,&hToken);
	isok=AdjustTokenPrivileges(hToken,0,&NewState,0x10,0,0);
	CloseHandle(hToken);

        // open tcp device		
        RtlInitUnicodeString(&dn,ch1);
        ofs.SecurityDescriptor = 0;
	ofs.ObjectName = &dn;
	ofs.Length =0x18;
	ofs.RootDirectory = 0;
	ofs.Attributes =0x40;
	ofs.SecurityQualityOfService =0;
	status=ZwOpenFile(&htcp,0x100000,&ofs,&ch3,3,0);

        // open udp device
        RtlInitUnicodeString(&dn,ch2);
 	ofs.ObjectName = &dn;
	status=ZwOpenFile(&hudp,0x100000,&ofs,&ch3,3,0);


	dwNumBytesRet = 0x10;

        // query for list of handles
        // 0x10 is not documented in winternl.h
	isok = (*NtQuerySystemInformation)(0x10,pdwHandleList,
                                             dwNumBytes,&dwNumBytesRet);

        // the actual handles in pdwHandleList need to be
        // closed 

	if (!isok) {

            // find the handle to the tcp device within this 
            // the handle list; match PID and device
            // why skip over the first one?

	    dwNumEntries = pdwHandleList[0];
	    pHandleInfo = (PHANDLEINFO)( pdwHandleList + 1 );

	    for (i = 0; i < dwNumEntries; i++ ) {
	 	if (pHandleInfo->dwPid == GetCurrentProcessId() && 
                    pHandleInfo->HndlOffset ==(int)htcp) {
	  	    tcpdnum = pHandleInfo;
		    break;
		}
		pHandleInfo++;
	     }

             // find the handle to the udp device within handle list
             // match PID and device
      	     pHandleInfo = (PHANDLEINFO)( pdwHandleList + 1 );
	     for (i = 0; i < dwNumEntries; i++ ) {
		 if (pHandleInfo->dwPid == GetCurrentProcessId() && 
                     pHandleInfo->HndlOffset == (int)hudp) {
		     udpdnum = pHandleInfo;
		     break;
		 }
	         pHandleInfo++;
	     }

	     ZeroMemory(buf1,0x70);
	     ZeroMemory(buf2,0x70);

             // read from the tcp device's memory?
	     readset1 = getmap(tcpdnum,pmaddr,h1,buf1);
	     if (readset1==0) {
		printf("win2kfport: map tcp fail\n");
		return 0;
	     }

             // read from udp device's memory?
	     readset2 = getmap(udpdnum,pmaddr,h1,buf2);
	     if (readset2==0) {
	        printf("win2kfport: map udp fail\n");
		return 0;
	     }
	
             // initialize OVERLAPPED struct for async I/O
             la.hEvent = CreateEvent(0,1,0,0);;
	     la.Internal = 0;
	     la.InternalHigh=0;
	     la.Offset = 0;
	     la.OffsetHigh = 0;

             // go through the handles, looking for tcp and udp
             // handles
             numHandles = 0;

             // for each tcp handle . . .
	     pHandleInfo = (PHANDLEINFO)( pdwHandleList + 1 );
	     for (i = 0; i < dwNumEntries; i++ ) {
	 	 ZeroMemory(buf3,0x70);

                 // if the handle is a tcp object type, read it

		 if (pHandleInfo->ObjType == tcpdnum->ObjType) {
		    readset3 = getmap(pHandleInfo,pmaddr,h1,buf3);
		    if (readset3==0) {
		       pHandleInfo++;
		       continue;
		    }

		    if (buf3[4]==buf1[4] && buf3[5]==buf1[5] && 
                        buf3[6]==buf1[6] && buf3[7]==buf1[7]) {

			if ((buf3[16]==1 || buf3[16]==2) && 
                             buf3[17]==0 && buf3[18]==0 && buf3[19]==0)	{
 	 	             hProc = OpenProcess(0x40,0,pHandleInfo->dwPid);
			     if (hProc == NULL) {
			        pHandleInfo++;
				continue;
			     }
                   	     DuplicateHandle(hProc,
                                             (HANDLE)pHandleInfo->HndlOffset,
                                             pmy,&myhand,0,0,2);
			     CloseHandle(hProc);
			     if (myhand==NULL) {
				 pHandleInfo++;
				 continue;
			     }
 	 	             ZeroMemory(out1,0x30);
			     ZeroMemory(out,0x38);
			     ZeroMemory(in,0x18);
			     ZeroMemory(in1,0x18);
			     in[0x10] = 4;
			     in1[0x10]=3;
			     if (buf3[16]==2) {
				p = 0;
                                // ioctl to the device (proc?)
				isok = DeviceIoControl(myhand,
                                                0x210012,&in,0x18,&out,
                                                 0x38,&p,&la);
				if (isok==FALSE) {
				   DBG("win2kport: DeviceIoctl of tcp hndle fail\n");
				   pHandleInfo++;
                  	           CloseHandle(myhand); // rdk
				   continue;
				}
			     }
			     isok = DeviceIoControl(myhand,0x210012,&in1,
                                                  0x18,&out1,0x30,&p,&la);
			     if (isok) {
                		port1 = out1[12];
				if (port1<0)
				    port1 = 256 + port1;
				port1 = port1*256;
				port2 = out1[13];
				if (port2<0)
 	 	 	   	   port1 = port1 + 256 + port2;
				else
				   port1 = port1 + port2;
 
                                enlarge_handle_cache();
                                handleCache[numHandles].pid = 
                                                 pHandleInfo->dwPid;
                                handleCache[numHandles].port = port1;
                                handleCache[numHandles].proto = 6;

				DBG2("TCP-> PID:%d PORT:%d\n",
                                     pHandleInfo->dwPid,port1);

                                numHandles++;
			     
                             } /* if DeviceIoControl */

			     CloseHandle(myhand);

		  	} /* buf3[16]==1 buf3[16]==2 . . .buf3[17] == 0 */

		    } /* buf3[4]==buf1[4] buf3[5]==buf1[5] ... */

		 } /* if objtype == tcpdnum->objtype */

		 pHandleInfo++;

	     } /* for each handle -- tcp round */

             // for each udp handle . . .
             pHandleInfo = (PHANDLEINFO)( pdwHandleList + 1 );
	     for (i = 0; i < dwNumEntries; i++ ) {
	  	 ZeroMemory(buf3,0x70);

                 // if the handle is a udp object type, read it 
 	   	 if (pHandleInfo->ObjType == udpdnum->ObjType) {
 	 	    readset3 = getmap(pHandleInfo,pmaddr,h1,buf3);
		    if (readset3==0) {
		       pHandleInfo++;
		       continue;
		     }

		     if (buf3[4]==buf2[4] && buf3[5]==buf2[5]&& 
                        buf3[6]==buf2[6] && buf3[7]==buf2[7]) {

		        if ((buf3[16]==1 || buf3[16]==2) && buf3[17]==0 && 
                             buf3[18]==0 && buf3[19]==0) {
			    hProc = OpenProcess(0x40,0,pHandleInfo->dwPid);
			    if (hProc==NULL) {
			       pHandleInfo++;
			       continue;
			    }
			    DuplicateHandle(hProc,
                                            (HANDLE)pHandleInfo->HndlOffset,
                                            pmy,&myhand,0,0,2);
			    CloseHandle(hProc);
			    if (myhand==NULL) {
			       pHandleInfo++;
			       continue;
			    }
			    ZeroMemory(out1,0x30);
			    ZeroMemory(out,0x38);
			    ZeroMemory(in,0x18);
			    ZeroMemory(in1,0x18);
			    in[0x10] = 4;
			    in1[0x10]=3;
			    if (buf3[16]==2) {
			       p = 0;
			       isok = DeviceIoControl(myhand,0x210012,&in,
                                                      0x18,&out,0x38,&p,&la);
			       if (isok==FALSE) {
	 	  	 	  CloseHandle(myhand); // rdk 
				  pHandleInfo++;
				  continue;
			       }
			    }
			    isok = DeviceIoControl(myhand,0x210012,&in1,0x18,
                                                   &out1,0x30,&p,&la);
			    if (isok) {
			       port1 = out1[12];
			       if (port1<0)
				  port1 = 256 + port1;
			       port1 = port1*256;
			       port2 = out1[13];
			       if(port2<0)
				 port1 = port1 + 256 + port2;
			       else
				 port1 = port1 + port2;
                               
			       DBG2("UDP-> PID:%d PORT:%d\n",
                                        pHandleInfo->dwPid,
                                       (unsigned short)port1);

                               enlarge_handle_cache();
                               handleCache[numHandles].pid = 
                                                pHandleInfo->dwPid;
                               handleCache[numHandles].port = port1;
                               handleCache[numHandles].proto = 17;

                               numHandles++;

			    }
			    CloseHandle(myhand);
			} /* buf3[18] */
		     } /* buf3[6] */
		 } /* if pHandleInfo->ObjType == udp */
	         pHandleInfo++;

	     } /* for i=0; i< dwNumEntries; */

             // close handles; we need to find the source
             // of leaks
             DBG("Closing handles\n");
 	     CloseHandle(la.hEvent);
	     CloseHandle(pmy);

	     //CloseHandle(h1);
             ZwClose(h1);

	     //CloseHandle(htcp);
	     //CloseHandle(hudp);
             // close using ZwClose() since that is what
             // ZwOpen says to use
             ZwClose(htcp);
             ZwClose(hudp);
	     UnmapViewOfFile(pmaddr);
        } /* if !isok */
	else {
	   printf("error open handlelist, %d\n",isok); 
        }

     } /* if status == 0 */
     else {
        printf("error open physcalmemory\n");
     }

     return 1;
}

/* ************************************************************* */

#endif /* WIN32 */
